{*********************************************************************
*
* Copyright (c) 1997-2000 Microsoft Corporation. All rights reserved
*
* File: D3DApp.h
*
* Desc: Application class for the Direct3D samples framework library.
*
*********************************************************************
* timeline :
*   12/31/2k -> alpha test SUCESSFULL !!!
*   12/30/2k -> alpha build, to rebuild the nice sdk samples :)
*
* by orthkon * orthkon@mail.com * http://mp3.com/orthkon
*
* changes:  1. sort algorithm (SortModes) was erroneous
*              - corrected to QuickSort (see unit "D3DUtil")
*           2. "\n" in message strings changed to "#13"
*           3. All methods used by an user application are declared in
*              TD3DApplication as virtual (see below). Override them with
*              specific user methods when deriving from TD3DApplication
*              (all former virtual function variables were removed
*              The main form has to do at least three things:
*                create the defined application         (in "OnCreate")
*                start it with "Init"                   (in "OnShow")
*                release it with "Cleanup3DEnvironment" (in "OnClose")
*           4. Delphi like variable names
* remarks: time from dwTime is in milliseconds (not in seconds)
*
*   J. Rathlev, Nov. 2001 (rathlev@physik.uni-kiel.de)
*
*********************************************************************}

{******************************************************************************
* Name: class TD3DApplication
* Desc: A base class for creating sample D3D8 applications. To create a
*       simple Direct3D application, simply derive this class into a
*       class (such as class MyD3DApplication) and override the following
*       virtual methods, as needed:
*          ConfirmDevice()           - To check the device for some minimum set
*                                      of capabilities
*          OneTimeSceneInit()        - To initialize app data (alloc mem, etc.)
*          InitDeviceObjects()       - To initialize the 3D scene objects
*          FrameMove()               - To animate the scene
*          Render()                  - To render the scene
*          RestoreDeviceObjects()    - To restore device-memory objects and state
*                                      after a device is created or resized
*          InvalidateDeviceObjects() - To be called when the device-dependent
*                                      objects are about to be lost.
*          DeleteDeviceObjects()     - To cleanup the 3D scene objects
*          FinalCleanup()            - To cleanup app data (for exitting the app)
*          MsgProc()                 - To handle Windows messages
******************************************************************************}
unit D3DApp;

interface

uses
  Windows, Messages, SysUtils, Classes, Forms, Dialogs,
  Direct3D8, D3DX8, D3DUtil, DXTimer;

const
  D3DAPPERR_NODIRECT3D          = HRESULT( $82000001 );
  D3DAPPERR_NOWINDOW            = HRESULT( $82000002 );
  D3DAPPERR_NOCOMPATIBLEDEVICES = HRESULT( $82000003 );
  D3DAPPERR_NOWINDOWABLEDEVICES = HRESULT( $82000004 );
  D3DAPPERR_NOHARDWAREDEVICE    = HRESULT( $82000005 );
  D3DAPPERR_HALNOTCOMPATIBLE    = HRESULT( $82000006 );
  D3DAPPERR_NOWINDOWEDHAL       = HRESULT( $82000007 );
  D3DAPPERR_NODESKTOPHAL        = HRESULT( $82000008 );
  D3DAPPERR_NOHALTHISMODE       = HRESULT( $82000009 );
  D3DAPPERR_NONZEROREFCOUNT     = HRESULT( $8200000a );
  D3DAPPERR_MEDIANOTFOUND       = HRESULT( $8200000b );
  D3DAPPERR_RESIZEFAILED        = HRESULT( $8200000c );

  MSG_NONE                      = $00000000;
  MSGERR_APPMUSTEXIT            = $00000001;
  MSGWARN_SWITCHEDTOREF         = $00000002;

  // constants form D3DRes.h
  IDI_MAIN_ICON                 = 101; // Application icon
  IDR_MAIN_ACCEL                = 113; // Keyboard accelerator
  IDR_MENU                      = 141; // Application menu
  IDR_POPUP                     = 142; // Popup menu
  IDD_SELECTDEVICE              = 144; // "Change Device" dialog box

  IDC_ADAPTER_COMBO             = 1002; // Adapter combobox for "SelectDevice" dlg
  IDC_DEVICE_COMBO              = 1000; // Device combobox for "SelectDevice" dlg
  IDC_FULLSCREENMODES_COMBO     = 1003; // Mode combobox for "SelectDevice" dlg
  IDC_MULTISAMPLE_COMBO         = 1005; // MultiSample combobox for "SelectDevice" dlg
  IDC_WINDOW                    = 1016; // Radio button for windowed-mode
  IDC_FULLSCREEN                = 1018; // Radio button for fullscreen-mode

  IDCHANGEDEVICE              = 40002; // Command to invoke "Change Device" dlg
  IDTOGGLEFULLSCREEN          = 40003; // Command to toggle fullscreen mode
  IDTOGGLESTART               = 40004; // Command to toggle frame animation
  IDSINGLESTEP                = 40005; // Command to single step frame animation
  IDEXIT                      = 40006; // Command to exit the application

type
  TD3DModeInfo = packed record
    Width              : DWORD;     // Screen width in this mode
    Height             : DWORD;     // Screen height in this mode
    Format             : TD3DFORMAT; // Pixel format in this mode
    dwBehavior         : DWORD;     // Hardware / Software / Mixed vertex processing
    DepthStencilFormat : TD3DFORMAT; // Which depth/stencil format to use with this mode
  end;

  TD3DDeviceInfo = packed record
    // Device data
    DeviceType      : TD3DDEVTYPE; // Reference, HAL, etc.
    d3dCaps         : TD3DCAPS8;   // Capabilities of this device
    strDesc         : PChar;      // Name of this device
    bCanDoWindowed  : Boolean;    // Whether this device can work in windowed mode

    // Modes for this device
    dwNumModes : DWORD;
    modes      : array[0..149] of TD3DModeInfo;

    // Current state
    dwCurrentMode   : DWORD;
    bWindowed       : Boolean;
    MultiSampleType : TD3DMULTISAMPLE_TYPE;
  end;

  TD3DAdapterInfo = packed record
    // Adapter data
    d3dAdapterIdentifier : TD3DADAPTER_IDENTIFIER8;
    d3ddmDesktop : TD3DDISPLAYMODE;      // Desktop display mode for this adapter

    // Devices for this adapter
    dwNumDevices : DWORD;
    devices : array[0..4] of TD3DDeviceInfo;

    // Current state
    dwCurrentDevice : DWORD;
  end;

  TD3DApplication = class( TComponent )
  private
    Owner : TForm;
    Timer : TDXTimer;
    // use in Render3DEnvironment
    dwLastTime, dwFrames : DWORD;
    // use in Pause
    dwAppPausedCount : DWORD;
    procedure Run( Sender: TObject );
    function BuildDeviceList : HRESULT;
    function ForceWindowed : HRESULT;
    function ToggleFullscreen : HRESULT;
    function Render3DEnvironment : HRESULT;
    function Resize3DEnvironment : HRESULT;
    function DisplayErrorMsg( hr : HRESULT; dwType : DWORD ) : HRESULT;
    function FindDepthStencilFormat( iAdapter : Word; DeviceType : TD3DDEVTYPE; TargetFormat : TD3DFORMAT; var pDepthStencilFormat : TD3DFORMAT ) : Boolean;
    function Initialize3DEnvironment : HRESULT;
    function AdjustWindowForChange : HRESULT;
    function SortModesCallback (const arg1,arg2) : boolean;
  protected
    constructor Create( AOwner : TComponent ); override;
    // virtual functions as placeholder for subclass
    function ConfirmDevice( Caps : TD3DCAPS8;
                dwBehavior : DWORD; Format : TD3DFORMAT ) : HRESULT; virtual;
    function OneTimeSceneInit : HRESULT; virtual;
    function InitDeviceObjects : HRESULT; virtual;
    function RestoreDeviceObjects : HRESULT; virtual;
    function FrameMove : HRESULT; virtual;
    function Render : HRESULT; virtual;
    function InvalidateDeviceObjects : HRESULT; virtual;
    function DeleteDeviceObjects : HRESULT; virtual;
    function FinalCleanup : HRESULT; virtual;
  public
    // Internal variables for the state of the app
    Adapters       : array[0..9] of TD3DAdapterInfo;
    dwNumAdapters  : DWORD;
    dwAdapter      : DWORD;
    bWindowed      : Boolean;
    bActive        : Boolean;
    bReady         : Boolean;

    // Internal variables used for timing
    bFrameMoving  : Boolean;
    bSingleStep   : Boolean;

    // Main objects used for creating and rendering the 3D scene
    d3dpp : TD3DPRESENT_PARAMETERS;         // Parameters for CreateDevice/Reset
    hWnd : HWND;              // The main app window
    hWndFocus : HWND;         // The D3D focus window (usually same as hWnd)
    D3D      : IDirect3D8;               // The main D3D object
    d3dDevice  : IDirect3DDevice8;       // The D3D rendering device
    d3dCaps : TD3DCAPS8;           // Caps for the device
    d3dsdBackBuffer : TD3DSURFACE_DESC;   // Surface desc of the backbuffer
    dwCreateFlags : DWORD;     // Indicate sw or hw vertex processing
    dwWindowStyle : DWORD;     // Saved window style for mode switches
    rcWindowBounds : TRECT;    // Saved window bounds for mode switches
    rcWindowClient : TRECT;    // Saved client area size for mode switches

    // Variables for timing
    dwTime : DWORD;             // Current time in milliseconds
    dwElapsedTime : DWORD;      // Time elapsed since last frame
    fFPS : Real;              // Instanteous frame rate
    strDeviceStats : String;// String to hold D3D device stats
    strFrameStats : String; // String to hold frame stats

    // Overridable variables for the app
    strWindowTitle : String;    // Title for the app's window
    bUseDepthBuffer : Boolean;   // Whether to autocreate depthbuffer
    dwMinDepthBits : DWORD;    // Minimum number of bits needed in depth buffer
    dwMinStencilBits : DWORD;  // Minimum number of bits needed in stencil buffer
    dwCreationWidth : DWORD;   // Width used to create window
    dwCreationHeight : DWORD;  // Height used to create window
    bShowCursorWhenFullscreen : Boolean; // Whether to show cursor when fullscreen
    function Init : HRESULT;
    procedure Pause( bPause : Boolean );
    procedure Cleanup3DEnvironment;
    procedure MsgProc( var Msg: TMsg; var Handled: Boolean );
  end;

implementation

const
  dwNumDeviceTypes : DWORD = 2;
  strDeviceDescs : array[0..1] of PChar = ( 'HAL', 'REF' );
  DeviceTypes : array[0..1] of TD3DDEVTYPE = ( D3DDEVTYPE_HAL, D3DDEVTYPE_REF );

function StrReplace( str, s_str, r_str : String ) : String;
var
  p1 : Integer;
begin
  repeat
    p1 := Pos( s_str, str );
    Result := Result + Copy( str, 1, p1 - 1 ) + r_str;
    Delete( str, 1, p1 + Length( s_str ) - 1 );
  until p1 = 0;
end;

// virtual ( property ) functions for subclass
function TD3DApplication.ConfirmDevice( Caps : TD3DCAPS8;
                             dwBehavior : DWORD; Format : TD3DFORMAT ) : HRESULT;
begin
  Result := S_OK;
  end;

function TD3DApplication.OneTimeSceneInit : HRESULT;
begin
  Result := S_OK;
  end;

function TD3DApplication.InitDeviceObjects : HRESULT;
begin
  Result := S_OK;
  end;

function TD3DApplication.RestoreDeviceObjects : HRESULT;
begin
  Result := S_OK;
  end;

function TD3DApplication.FrameMove : HRESULT;
begin
  Result := S_OK;
  end;

function TD3DApplication.Render : HRESULT;
begin
  Result := S_OK;
  end;

function TD3DApplication.InvalidateDeviceObjects : HRESULT;
begin
  Result := S_OK;
  end;

function TD3DApplication.DeleteDeviceObjects : HRESULT;
begin
  Result := S_OK;
  end;

function TD3DApplication.FinalCleanup : HRESULT;
begin
  Result := S_OK;
  end;

constructor TD3DApplication.Create( AOwner : TComponent );
begin
  Owner := TForm( AOwner );

  dwNumAdapters     := 0;
  dwAdapter         := 0;
  D3D               := nil;
  d3dDevice         := nil;
  hWnd              := 0;
  hWndFocus         := 0;
  bActive           := FALSE;
  bReady            := FALSE;
  dwCreateFlags     := 0;

  bFrameMoving      := true;
  bSingleStep       := FALSE;
  fFPS              := 0.0;
  strDeviceStats    := '';
  strFrameStats     := '';

  strWindowTitle    := 'D3D8 Application';
  dwCreationWidth   := 400;
  dwCreationHeight  := 300;
  bUseDepthBuffer   := FALSE;
  dwMinDepthBits    := 16;
  dwMinStencilBits  := 0;
  bShowCursorWhenFullscreen := false;

  // static vars
  dwFrames := 0;
  dwElapsedTime := 0;
  dwLastTime := 0;
  dwAppPausedCount := 0;

end;

function TD3DApplication.Init : HRESULT;
begin
  Result := E_FAIL;

  // Create the Direct3D object
  D3D := Direct3DCreate8( D3D_SDK_VERSION );
  if D3D = nil then begin
    DisplayErrorMsg( D3DAPPERR_NODIRECT3D, MSGERR_APPMUSTEXIT );
    Exit;
  end;

  // Build a list of Direct3D adapters, modes and devices. The
  // ConfirmDevice() callback is used to confirm that only devices that
  // meet the app's requirements are considered.
  Result := BuildDeviceList;
  if Failed( Result ) then begin
    D3D := nil;
    DisplayErrorMsg( Result, MSGERR_APPMUSTEXIT );
    Exit;
  end;

  hWnd := Owner.Handle;

  // The focus window can be a specified to be a different window than the
  // device window.  If not, use the device window as the focus window.
  if hWndFocus = 0 then hWndFocus := hWnd;

  // Save window properties
  dwWindowStyle := GetWindowLong( hWnd, GWL_STYLE );
  GetWindowRect( hWnd, rcWindowBounds );
  GetClientRect( hWnd, rcWindowClient );

  // Initialize the app's custom scene stuff
  Result := OneTimeSceneInit;
  if Failed( Result ) then begin
    D3D := nil;
    Result := DisplayErrorMsg( Result, MSGERR_APPMUSTEXIT );
    Exit;
  end;

  // Initialize the 3D environment for the app
  Result := Initialize3DEnvironment;
  if Failed( Result ) then begin
    D3D := nil;
    Result := DisplayErrorMsg( Result, MSGERR_APPMUSTEXIT );
    Exit;
  end;

  // The app is ready to go
  bReady := TRUE;

  // set timer routine
  Timer := TDXTimer.Create( Self );
  Timer.OnTimer := Run;
  Timer.Interval := 1;

  Application.OnMessage := MsgProc;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: SortModesCallback()
// Desc: Callback function for sorting display modes (used by BuildDeviceList).
//-----------------------------------------------------------------------------
function TD3DApplication.SortModesCallback (const arg1,arg2) : boolean;
begin
  Result:=(TD3DDISPLAYMODE(arg1).Format>TD3DDISPLAYMODE(arg2).Format)
           or (TD3DDISPLAYMODE(arg1).Width>TD3DDISPLAYMODE(arg2).Width)
           or (TD3DDISPLAYMODE(arg1).Height>TD3DDISPLAYMODE(arg2).Height);
  end;

//-----------------------------------------------------------------------------
// Name: BuildDeviceList()
// Desc:
//-----------------------------------------------------------------------------
function TD3DApplication.BuildDeviceList : HRESULT;
var
  bHALExists, bHALIsWindowedCompatible,
  bHALIsDesktopCompatible, bHALIsSampleCompatible : Boolean;
  iDevice, iMode, iAdapter : Word;
  pAdapter : ^TD3DAdapterInfo;
  modes : array[0..119] of TD3DDISPLAYMODE;
  formats : array[0..19] of TD3DFORMAT;
  dwNumFormats, dwNumModes, dwNumAdapterModes : DWORD;
  DisplayMode : TD3DDISPLAYMODE ;
  pDevice : ^TD3DDeviceInfo;
  d, a, f, m : DWORD;
  bFormatConfirmed : array[0..19] of Boolean;
  dwBehavior : array[0..19] of DWORD;
  fmtDepthStencil : array[0..19] of TD3DFORMAT;
begin
  bHALExists := False;
  bHALIsWindowedCompatible := False;
  bHALIsDesktopCompatible := False;
  bHALIsSampleCompatible := False;

  // Loop through all the adapters on the system (usually, there's just one
  // unless more than one graphics card is present).
  for iAdapter := 0 to D3D.GetAdapterCount - 1 do begin

    // Fill in adapter info
    pAdapter := @Adapters[dwNumAdapters];
    D3D.GetAdapterIdentifier( iAdapter, 0, pAdapter^.d3dAdapterIdentifier );
    D3D.GetAdapterDisplayMode( iAdapter, pAdapter^.d3ddmDesktop );
    pAdapter^.dwNumDevices    := 0;
    pAdapter^.dwCurrentDevice := 0;

    // Enumerate all display modes on this adapter
    dwNumFormats := 0;
    dwNumModes := 0;
    dwNumAdapterModes := D3D.GetAdapterModeCount( iAdapter );

    // Add the adapter's current desktop format to the list of formats
    formats[dwNumFormats] := pAdapter^.d3ddmDesktop.Format;
    inc( dwNumFormats );

    for iMode := 0 to dwNumAdapterModes - 1 do begin
      // Get the display mode attributes
      D3D.EnumAdapterModes( iAdapter, iMode, DisplayMode );

      // Filter out low-resolution modes
      if ( DisplayMode.Width  < 640 ) or ( DisplayMode.Height < 400 ) then
        continue;

      // Check if the mode already exists (to filter out refresh rates)
      m := 0;
      while m < dwNumModes do begin
        if ( modes[m].Width = DisplayMode.Width  ) and
           ( modes[m].Height = DisplayMode.Height ) and
           ( modes[m].Format = DisplayMode.Format ) then break;
        inc( m );
      end;

      // If we found a new mode, add it to the list of modes
      if m = dwNumModes then begin
        modes[dwNumModes].Width       := DisplayMode.Width;
        modes[dwNumModes].Height      := DisplayMode.Height;
        modes[dwNumModes].Format      := DisplayMode.Format;
        modes[dwNumModes].RefreshRate := 0;
        inc( dwNumModes );

        // Check if the mode's format already exists
        for f := 0 to dwNumFormats - 1 do
        if DisplayMode.Format = formats[f] then break;

        // If the format is new, add it to the list
        if f = dwNumFormats then begin
          formats[dwNumFormats] := DisplayMode.Format;
          inc( dwNumFormats );
        end;

      end;

    end;

    // Sort the list of display modes (by format, then width, then height)
//    SortModes( modes, dwNumModes );
    QuickSort(modes,dwNumModes,sizeof(D3DDISPLAYMODE),SortModesCallback);

    // Add devices to adapter
    for iDevice := 0 to dwNumDeviceTypes - 1 do begin

      // Fill in device info
      pDevice                  := @pAdapter^.devices[pAdapter^.dwNumDevices];
      pDevice^.DeviceType      := DeviceTypes[iDevice];
      D3D.GetDeviceCaps( iAdapter, DeviceTypes[iDevice], pDevice^.d3dCaps );
      pDevice^.strDesc         := strDeviceDescs[iDevice];
      pDevice^.dwNumModes      := 0;
      pDevice^.dwCurrentMode   := 0;
      pDevice^.bCanDoWindowed  := FALSE;
      pDevice^.bWindowed       := FALSE;
      pDevice^.MultiSampleType := D3DMULTISAMPLE_NONE;

      // Examine each format supported by the adapter to see if it will
      // work with this device and meets the needs of the application.
      for f := 0 to dwNumFormats - 1 do begin
        bFormatConfirmed[f] := FALSE;
        fmtDepthStencil[f]  := D3DFMT_UNKNOWN;

        // Skip formats that cannot be used as render targets on this device
        if Failed( D3D.CheckDeviceType
          ( iAdapter, pDevice^.DeviceType, formats[f], formats[f], False ) ) then
            continue;

        if pDevice^.DeviceType = D3DDEVTYPE_HAL then begin
          // This system has a HAL device
          bHALExists := TRUE;

          if ( pDevice^.d3dCaps.Caps2 and D3DCAPS2_CANRENDERWINDOWED ) = D3DCAPS2_CANRENDERWINDOWED then begin
            // HAL can run in a window for some mode
            bHALIsWindowedCompatible := TRUE;

            if f = 0 then begin
              // HAL can run in a window for the current desktop mode
              bHALIsDesktopCompatible := TRUE;
            end;

          end;

        end;

        // Confirm the device/format for HW vertex processing
        if ( pDevice^.d3dCaps.DevCaps and D3DDEVCAPS_HWTRANSFORMANDLIGHT ) = D3DDEVCAPS_HWTRANSFORMANDLIGHT then begin

          if ( pDevice^.d3dCaps.DevCaps and D3DDEVCAPS_PUREDEVICE ) = D3DDEVCAPS_PUREDEVICE then begin
            dwBehavior[f] := D3DCREATE_HARDWARE_VERTEXPROCESSING or D3DCREATE_PUREDEVICE;

            if SUCCEEDED(ConfirmDevice
              ( pDevice^.d3dCaps, dwBehavior[f], formats[f] ) ) then
                bFormatConfirmed[f] := TRUE;
          end;

          if FALSE = bFormatConfirmed[f] then begin
            dwBehavior[f] := D3DCREATE_HARDWARE_VERTEXPROCESSING;

            if SUCCEEDED(ConfirmDevice
              ( pDevice^.d3dCaps, dwBehavior[f], formats[f] ) ) then
                bFormatConfirmed[f] := TRUE;
          end;

          if FALSE = bFormatConfirmed[f] then begin
            dwBehavior[f] := D3DCREATE_MIXED_VERTEXPROCESSING;

            if SUCCEEDED(ConfirmDevice
              ( pDevice^.d3dCaps, dwBehavior[f], formats[f] ) ) then
                bFormatConfirmed[f] := TRUE;
          end;

        end;

        // Confirm the device/format for SW vertex processing
        if FALSE = bFormatConfirmed[f] then begin
          dwBehavior[f] := D3DCREATE_SOFTWARE_VERTEXPROCESSING;

          if SUCCEEDED(ConfirmDevice
            ( pDevice^.d3dCaps, dwBehavior[f], formats[f] ) ) then
              bFormatConfirmed[f] := TRUE;
        end;

        // Find a suitable depth/stencil buffer format for this device/format
        if bFormatConfirmed[f] and bUseDepthBuffer then begin
          if not FindDepthStencilFormat( iAdapter, pDevice.DeviceType
          , formats[f], fmtDepthStencil[f] ) then
            bFormatConfirmed[f] := FALSE;
        end;

      end; // dwNumFormats

      // Add all enumerated display modes with confirmed formats to the
      // device's list of valid modes
      m := 0;
      while m < dwNumModes do begin
        f := 0;
        while f < dwNumFormats do begin
          if modes[m].Format = formats[f] then
          if bFormatConfirmed[f] = TRUE then begin

            // Add this mode to the device's list of valid modes
            pDevice^.modes[pDevice^.dwNumModes].Width      := modes[m].Width;
            pDevice^.modes[pDevice^.dwNumModes].Height     := modes[m].Height;
            pDevice^.modes[pDevice^.dwNumModes].Format     := modes[m].Format;
            pDevice^.modes[pDevice^.dwNumModes].dwBehavior := dwBehavior[f];
            pDevice^.modes[pDevice^.dwNumModes].DepthStencilFormat := fmtDepthStencil[f];
            inc( pDevice^.dwNumModes );

            if pDevice^.DeviceType = D3DDEVTYPE_HAL then
              bHALIsSampleCompatible := TRUE;
          end;
          inc( f );
        end;
        inc( m );
      end;

      // Select any 640x480 mode for default (but prefer a 16-bit mode)
      m := 0;
      while m < pDevice^.dwNumModes do begin
        if ( pDevice^.modes[m].Width = 640 ) and ( pDevice^.modes[m].Height = 480 ) then begin
          pDevice^.dwCurrentMode := m;
          if ( pDevice^.modes[m].Format = D3DFMT_R5G6B5 )
          or ( pDevice^.modes[m].Format = D3DFMT_X1R5G5B5 )
          or ( pDevice^.modes[m].Format = D3DFMT_A1R5G5B5 ) then break;
        end;
        inc( m );
      end;

      // Check if the device is compatible with the desktop display mode
      // (which was added initially as formats[0])
      if bFormatConfirmed[0]
      and ( ( pDevice^.d3dCaps.Caps2 and D3DCAPS2_CANRENDERWINDOWED )
      = D3DCAPS2_CANRENDERWINDOWED ) then begin
        pDevice^.bCanDoWindowed := TRUE;
        pDevice^.bWindowed      := TRUE;
      end;

      // If valid modes were found, keep this device
      if pDevice^.dwNumModes > 0 then inc( pAdapter^.dwNumDevices );

    end; // iDevice

    // If valid devices were found, keep this adapter
    if pAdapter^.dwNumDevices > 0 then inc( dwNumAdapters );

  end;

  // Return an error if no compatible devices were found
  if 0 = dwNumAdapters then begin
    Result := HRESULT( D3DAPPERR_NOCOMPATIBLEDEVICES );
    Exit;
  end;

  // Pick a default device that can render into a window
  // (This code assumes that the HAL device comes before the REF
  // device in the device array).
  a := 0;
  while a < dwNumAdapters do begin
    d := 0;
    while d < Adapters[a].dwNumDevices do begin
      if Adapters[a].devices[d].bWindowed then begin

        Adapters[a].dwCurrentDevice := d;
        dwAdapter := a;
        bWindowed := TRUE;

        // Display a warning message
        if Adapters[a].devices[d].DeviceType = D3DDEVTYPE_REF then begin
          if not bHALExists then
            DisplayErrorMsg( D3DAPPERR_NOHARDWAREDEVICE, MSGWARN_SWITCHEDTOREF )
          else if not bHALIsSampleCompatible then
            DisplayErrorMsg( D3DAPPERR_HALNOTCOMPATIBLE, MSGWARN_SWITCHEDTOREF )
          else if not bHALIsWindowedCompatible then
            DisplayErrorMsg( D3DAPPERR_NOWINDOWEDHAL, MSGWARN_SWITCHEDTOREF )
          else if not bHALIsDesktopCompatible then
            DisplayErrorMsg( D3DAPPERR_NODESKTOPHAL, MSGWARN_SWITCHEDTOREF )
          else // HAL is desktop compatible, but not sample compatible
            DisplayErrorMsg( D3DAPPERR_NOHALTHISMODE, MSGWARN_SWITCHEDTOREF );
        end;
        Result := S_OK;
        Exit;
      end;
      inc( d );
    end;
    inc( a );
  end;
  Result := HRESULT( D3DAPPERR_NOWINDOWABLEDEVICES );
end;

//-----------------------------------------------------------------------------
// Name: FindDepthStencilFormat()
// Desc: Finds a depth/stencil format for the given device that is compatible
//       with the render target format and meets the needs of the app.
//-----------------------------------------------------------------------------
function TD3DApplication.FindDepthStencilFormat( iAdapter : Word; DeviceType : TD3DDEVTYPE;
    TargetFormat : TD3DFORMAT; var pDepthStencilFormat : TD3DFORMAT ) : Boolean;
begin
  Result := True;

  if ( dwMinDepthBits <= 16 ) and ( dwMinStencilBits = 0 ) then
  if SUCCEEDED( D3D.CheckDeviceFormat( iAdapter, DeviceType
  , TargetFormat, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE
  , D3DFMT_D16 ) ) then
  if SUCCEEDED( D3D.CheckDepthStencilMatch( iAdapter, DeviceType
  , TargetFormat, TargetFormat, D3DFMT_D16 ) ) then begin
    pDepthStencilFormat := D3DFMT_D16;
    Exit;
  end;

  if ( dwMinDepthBits <= 15 ) and ( dwMinStencilBits <= 1 ) then
  if SUCCEEDED( D3D.CheckDeviceFormat( iAdapter, DeviceType
  , TargetFormat, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE
  , D3DFMT_D15S1 ) ) then
  if SUCCEEDED( D3D.CheckDepthStencilMatch( iAdapter, DeviceType
  , TargetFormat, TargetFormat, D3DFMT_D15S1 ) ) then begin
    pDepthStencilFormat := D3DFMT_D15S1;
    Exit;
  end;

  if ( dwMinDepthBits <= 24 ) and ( dwMinStencilBits = 0 ) then
  if SUCCEEDED( D3D.CheckDeviceFormat( iAdapter, DeviceType
  , TargetFormat, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE
  , D3DFMT_D24X8 ) ) then
  if SUCCEEDED( D3D.CheckDepthStencilMatch( iAdapter, DeviceType
  , TargetFormat, TargetFormat, D3DFMT_D24X8 ) ) then begin
    pDepthStencilFormat := D3DFMT_D24X8;
    Exit;
  end;

  if ( dwMinDepthBits <= 24 ) and ( dwMinStencilBits <= 8 ) then
  if SUCCEEDED( D3D.CheckDeviceFormat( iAdapter, DeviceType
  , TargetFormat, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE
  , D3DFMT_D24S8 ) ) then
  if SUCCEEDED( D3D.CheckDepthStencilMatch( iAdapter, DeviceType
  , TargetFormat, TargetFormat, D3DFMT_D24S8 ) ) then begin
    pDepthStencilFormat := D3DFMT_D24S8;
    Exit;
  end;

  if ( dwMinDepthBits <= 24 ) and ( dwMinStencilBits <= 4 ) then
  if SUCCEEDED( D3D.CheckDeviceFormat( iAdapter, DeviceType
    , TargetFormat, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE
    , D3DFMT_D24X4S4 ) ) then
  if SUCCEEDED( D3D.CheckDepthStencilMatch( iAdapter, DeviceType
    , TargetFormat, TargetFormat, D3DFMT_D24X4S4 ) ) then begin
    pDepthStencilFormat := D3DFMT_D24X4S4;
    Exit;
  end;

  if ( dwMinDepthBits <= 32 ) and ( dwMinStencilBits = 0 ) then
  if SUCCEEDED( D3D.CheckDeviceFormat( iAdapter, DeviceType
  , TargetFormat, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE
  , D3DFMT_D32 ) ) then
  if SUCCEEDED( D3D.CheckDepthStencilMatch( iAdapter, DeviceType
  , TargetFormat, TargetFormat, D3DFMT_D32 ) ) then begin
    pDepthStencilFormat := D3DFMT_D32;
    Exit;
  end;

  Result := False;
end;

procedure TD3DApplication.MsgProc( var Msg: TMsg; var Handled: Boolean );
var
  pmmi : ^MINMAXINFO;
  rcClientOld : TRECT;
  hr : HRESULT;
  ptCursor : TPOINT;
begin
  case Msg.message of
    WM_PAINT: begin
      // Handle paint messages when the app is not ready
      if ( d3dDevice <> nil ) and not bReady then
      if bWindowed then
        d3dDevice.Present( nil, nil, 0, nil );
    end;
    WM_GETMINMAXINFO: begin
      pmmi := Pointer( Msg.lParam );
      pmmi^.ptMinTrackSize.x := 100;
      pmmi^.ptMinTrackSize.y := 100;
    end;
    WM_ENTERSIZEMOVE: begin
      // Halt frame movement while the app is sizing or moving
      if bFrameMoving then Timer.Enabled := False;
    end;
    WM_SIZE: begin
      ShowMessage( 'WEXITSIZEMOVE' );
      // Check to see if we are losing our window...
      if ( SIZE_MAXHIDE = Msg.wParam ) or ( SIZE_MINIMIZED = Msg.wParam ) then
        bActive := FALSE
      else
        bActive := TRUE;
    end;
    WM_EXITSIZEMOVE: begin
      if bFrameMoving then Timer.Enabled := True;
      if bActive and bWindowed then begin
        rcClientOld := rcWindowClient;

        // Update window properties
        GetWindowRect( hWnd, rcWindowBounds );
        GetClientRect( hWnd, rcWindowClient );

        if( rcClientOld.right - rcClientOld.left <> rcWindowClient.right - rcWindowClient.left ) or
          ( rcClientOld.bottom - rcClientOld.top <> rcWindowClient.bottom - rcWindowClient.top ) then begin
          // A new window size will require a new backbuffer
          // size, so the 3D structures must be changed accordingly.
          bReady := FALSE;

          d3dpp.BackBufferWidth  := rcWindowClient.right - rcWindowClient.left;
          d3dpp.BackBufferHeight := rcWindowClient.bottom - rcWindowClient.top;

          // Resize the 3D environment
          hr := Resize3DEnvironment;
          if Failed( hr ) then begin
            DisplayErrorMsg( D3DAPPERR_RESIZEFAILED, MSGERR_APPMUSTEXIT );
            Exit;
          end;

          bReady := TRUE;
        end;
      end;
    end;
    WM_SETCURSOR:
      // Turn off Windows cursor in fullscreen mode
      if bActive and bReady and not bWindowed then begin
        SetCursor( 0 );
        if bShowCursorWhenFullscreen then
          d3dDevice.ShowCursor( TRUE );
          Handled := TRUE; // prevent Windows from setting cursor to window class cursor
        end;
    WM_MOUSEMOVE:
      if bActive and bReady and ( d3dDevice <> nil ) then begin
        GetCursorPos( ptCursor );
//        ScreenToClient( hWnd, ptCursor );
        d3dDevice.SetCursorPosition( ptCursor.x, ptCursor.y, 0 );
      end;
    WM_ENTERMENULOOP:
      // Pause the app when menus are displayed
      Pause( TRUE );
    WM_EXITMENULOOP:
      Pause(FALSE);
    WM_CONTEXTMENU: begin
      // No context menus allowed in fullscreen mode
      if bWindowed then begin
        // Handle the app's context menu (via right mouse click)
        //TrackPopupMenuEx( GetSubMenu( LoadMenu( 0, MAKEINTRESOURCE(IDR_POPUP) ), 0 ),
                    //TPVERTICAL, LOWORD(Msg.lParam), HIWORD(Msg.lParam), Msg.hWnd, nil );
      end;
    end;
    WM_NCHITTEST:
      // Prevent the user from selecting the menu in fullscreen mode
      if not bWindowed then Handled := True;
    WM_SYSCOMMAND: begin
      // Prevent moving/sizing and power loss in fullscreen mode
      case Msg.wParam of
        SC_MOVE, SC_SIZE, SC_MAXIMIZE, SC_KEYMENU, SC_MONITORPOWER:
          if not bWindowed then begin
            Handled := True;
            Exit;
          end;
      end;
    end;
    WM_COMMAND: begin
      case LOWORD( Msg.wParam ) of
        IDTOGGLESTART: begin
          // Toggle frame movement
          bFrameMoving := not bFrameMoving;
          if bFrameMoving then Timer.Enabled := True else Timer.Enabled := False;
        end;
        IDSINGLESTEP: begin
          // Single-step frame movement
          if bFrameMoving then Timer.Enabled := False;
          //else DXUtil_Timer( TIMER_ADVANCE );
          bFrameMoving := FALSE;
          bSingleStep  := TRUE;
        end;
        IDCHANGEDEVICE: begin
          // Prompt the user to select a new device or mode
          if bActive and bReady then begin
            Pause( True );
            Handled := True;
            //hr := UserSelectNewDevice;
            hr := S_OK;
            if Failed( hr ) then Exit;
            Pause( False );
            Exit;
          end;
        end;
        IDTOGGLEFULLSCREEN: begin
          // Toggle the fullscreen/window mode
          if bActive and bReady then begin
            Handled := True;
            if Failed( ToggleFullscreen ) then
              DisplayErrorMsg( D3DAPPERR_RESIZEFAILED, MSGERR_APPMUSTEXIT );
            Exit;
          end;
        end;
        IDEXIT: begin
          // Recieved key/menu command to exit app
          SendMessage( Msg.hWnd, WM_CLOSE, 0, 0 );
          Handled := True;
          Exit;
        end;
      end;
    end;
  end;
  Handled := False;
end;

//-----------------------------------------------------------------------------
// Name: Initialize3DEnvironment()
// Desc:
//-----------------------------------------------------------------------------
function TD3DApplication.Initialize3DEnvironment : HRESULT;
var
  pAdapterInfo : ^TD3DAdapterInfo;
  pDeviceInfo : ^TD3DDeviceInfo;
  pModeInfo : ^TD3DModeInfo;
  BackBuffer : IDIRECT3DSURFACE8;
  hCursor : Windows.HCURSOR;
  i : Word;
begin
  pAdapterInfo := @Adapters[dwAdapter];
  pDeviceInfo  := @pAdapterInfo^.devices[pAdapterInfo^.dwCurrentDevice];
  pModeInfo    := @pDeviceInfo^.modes[pDeviceInfo^.dwCurrentMode];

  // Prepare window for possible windowed/fullscreen change
  AdjustWindowForChange;

  // Set up the presentation parameters
  ZeroMemory( @d3dpp, sizeof( d3dpp ) );
  d3dpp.Windowed               := pDeviceInfo^.bWindowed;
  d3dpp.BackBufferCount        := 1;
  d3dpp.MultiSampleType        := pDeviceInfo^.MultiSampleType;
  d3dpp.SwapEffect             := D3DSWAPEFFECT_DISCARD;
  d3dpp.EnableAutoDepthStencil := bUseDepthBuffer;
  d3dpp.AutoDepthStencilFormat := pModeInfo^.DepthStencilFormat;
  d3dpp.hDeviceWindow          := hWnd;

  if bWindowed then begin
    d3dpp.BackBufferWidth  := rcWindowClient.right - rcWindowClient.left;
    d3dpp.BackBufferHeight := rcWindowClient.bottom - rcWindowClient.top;
    d3dpp.BackBufferFormat := pAdapterInfo^.d3ddmDesktop.Format;
  end
  else begin
    d3dpp.BackBufferWidth  := pModeInfo^.Width;
    d3dpp.BackBufferHeight := pModeInfo^.Height;
    d3dpp.BackBufferFormat := pModeInfo^.Format;
  end;

  // Create the device
  Result := D3D.CreateDevice( dwAdapter, pDeviceInfo^.DeviceType
            , hWndFocus, pModeInfo^.dwBehavior, d3dpp, d3dDevice );
  if SUCCEEDED( Result ) then begin
    // When moving from fullscreen to windowed mode, it is important to
    // adjust the window size after recreating the device rather than
    // beforehand to ensure that you get the window size you want.  For
    // example, when switching from 640x480 fullscreen to windowed with
    // a 1000x600 window on a 1024x768 desktop, it is impossible to set
    // the window size to 1000x600 until after the display mode has
    // changed to 1024x768, because windows cannot be larger than the
    // desktop.
    if bWindowed then begin
      SetWindowPos( hWnd, HWND_NOTOPMOST,
                  rcWindowBounds.left, rcWindowBounds.top,
                  ( rcWindowBounds.right - rcWindowBounds.left ),
                  ( rcWindowBounds.bottom - rcWindowBounds.top ),
                  SWP_SHOWWINDOW );
    end;

    // Store device Caps
    d3dDevice.GetDeviceCaps( d3dCaps );
    dwCreateFlags := pModeInfo^.dwBehavior;

    // Store device description
    if pDeviceInfo^.DeviceType = D3DDEVTYPE_REF then
      strDeviceStats := 'REF'
    else if pDeviceInfo^.DeviceType = D3DDEVTYPE_HAL then
      strDeviceStats := 'HAL'
    else if pDeviceInfo^.DeviceType = D3DDEVTYPE_SW then
      strDeviceStats := 'SW';

    if ( ( pModeInfo^.dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING ) = D3DCREATE_HARDWARE_VERTEXPROCESSING )
    and ( ( pModeInfo^.dwBehavior and D3DCREATE_PUREDEVICE ) = D3DCREATE_PUREDEVICE ) then begin
      if pDeviceInfo^.DeviceType = D3DDEVTYPE_HAL then
        strDeviceStats := strDeviceStats + ' (pure hw vp)'
      else
        strDeviceStats := strDeviceStats + ' (simulated pure hw vp)';
    end
    else if ( pModeInfo^.dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING ) = D3DCREATE_HARDWARE_VERTEXPROCESSING then begin
      if pDeviceInfo^.DeviceType = D3DDEVTYPE_HAL then
        strDeviceStats := strDeviceStats + ' (hw vp)'
      else
        strDeviceStats := strDeviceStats + ' (simulated hw vp)';
    end
    else if ( pModeInfo^.dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING ) = D3DCREATE_MIXED_VERTEXPROCESSING then begin
      if pDeviceInfo^.DeviceType = D3DDEVTYPE_HAL then
        strDeviceStats := strDeviceStats + ' (mixed vp)'
      else
        strDeviceStats := strDeviceStats + ' (simulated mixed vp)';
    end
    else if ( pModeInfo^.dwBehavior and D3DCREATE_SOFTWARE_VERTEXPROCESSING ) = D3DCREATE_SOFTWARE_VERTEXPROCESSING then begin
      strDeviceStats := strDeviceStats + ' (sw vp)';
    end;

    if pDeviceInfo^.DeviceType = D3DDEVTYPE_HAL then begin
      strDeviceStats := strDeviceStats + ': ';
      strDeviceStats := strDeviceStats + pAdapterInfo^.d3dAdapterIdentifier.Description;
    end;

    // Store render target surface desc
    d3dDevice.GetBackBuffer( 0, D3DBACKBUFFER_TYPE_MONO, BackBuffer );
    BackBuffer.GetDesc( d3dsdBackBuffer );
    //BackBuffer._Release;
    BackBuffer := nil;

    // Set up the fullscreen cursor
    if bShowCursorWhenFullscreen and not bWindowed then begin
      hCursor := LoadCursor( 0, IDC_ARROW	);
      D3DUtil_SetDeviceCursor( d3dDevice, hCursor );
      d3dDevice.ShowCursor( TRUE );
    end;

    // Initialize the app's device-dependent objects
    Result:=InitDeviceObjects;
    if SUCCEEDED( Result ) then begin
      Result :=RestoreDeviceObjects;
      if SUCCEEDED( Result ) then begin
        bActive := TRUE;
        Result := S_OK;
        Exit;
      end;
    end;

    if Result=D3DAPPERR_MEDIANOTFOUND then exit;

    // Cleanup before we try again
    InvalidateDeviceObjects;
    DeleteDeviceObjects;
    d3dDevice := nil;
  end;

  // If that failed, fall back to the reference rasterizer
  if pDeviceInfo^.DeviceType = D3DDEVTYPE_HAL then begin
    // Let the user know we are switching from HAL to the reference rasterizer
    DisplayErrorMsg( Result, MSGWARN_SWITCHEDTOREF );

    // Select the default adapter
    dwAdapter := 0;
    pAdapterInfo := @Adapters[dwAdapter];

    // Look for a software device
    i := 0;
    while i < pAdapterInfo^.dwNumDevices do begin
      if pAdapterInfo^.devices[i].DeviceType = D3DDEVTYPE_REF then begin
        pAdapterInfo^.dwCurrentDevice := i;
        pDeviceInfo := @pAdapterInfo^.devices[i];
        bWindowed := pDeviceInfo^.bWindowed;
        break;
      end;
      inc( i );
    end;

    // Try again, this time with the reference rasterizer
    if pAdapterInfo^.devices[pAdapterInfo^.dwCurrentDevice].DeviceType = D3DDEVTYPE_REF then begin
      Result := Initialize3DEnvironment;
    end;

  end;

end;

//-----------------------------------------------------------------------------
// Name: Resize3DEnvironment
// Desc:
//-----------------------------------------------------------------------------
function TD3DApplication.Resize3DEnvironment : HRESULT;
var
  BackBuffer : IDIRECT3DSURFACE8;
  hCursor : Windows.HCURSOR;
begin
  // Release all vidmem objects
  Result := InvalidateDeviceObjects;
  if Failed( Result ) then Exit;

  // Reset the device
  Result := d3dDevice.Reset( d3dpp );
  if Failed( Result ) then Exit;

  // Store render target surface desc
  d3dDevice.GetBackBuffer( 0, D3DBACKBUFFER_TYPE_MONO, BackBuffer );
  BackBuffer.GetDesc( d3dsdBackBuffer );
  //BackBuffer._Release;
  BackBuffer := nil;

  // Set up the fullscreen cursor
  if bShowCursorWhenFullscreen and not bWindowed then begin
    hCursor := LoadCursor( 0, IDC_ARROW );
    D3DUtil_SetDeviceCursor( d3dDevice, hCursor );
    d3dDevice.ShowCursor( TRUE );
  end;

  // Initialize the app's device-dependent objects
  Result := RestoreDeviceObjects;
  if Failed( Result ) then Exit;

  // If the app is paused, trigger the rendering of the current frame
  if FALSE = bFrameMoving then begin
    bSingleStep := TRUE;
  end;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ToggleFullScreen()
// Desc: Called when user toggles between fullscreen mode and windowed mode
//-----------------------------------------------------------------------------
function TD3DApplication.ToggleFullscreen : HRESULT;
var
  pAdapterInfo : ^TD3DAdapterInfo;
  pDeviceInfo : ^TD3DDeviceInfo;
  pModeInfo : ^TD3DModeInfo;
begin
  // Get access to current adapter, device, and mode
  pAdapterInfo := @Adapters[dwAdapter];
  pDeviceInfo  := @pAdapterInfo^.devices[pAdapterInfo^.dwCurrentDevice];
  pModeInfo    := @pDeviceInfo^.modes[pDeviceInfo^.dwCurrentMode];

  // Need device change if going windowed and the current device
  // can only be fullscreen
  if not bWindowed and not pDeviceInfo^.bCanDoWindowed then begin
    Result := ForceWindowed;
    Exit;
  end;

  bReady := FALSE;

  // Toggle the windowed state
  bWindowed := not bWindowed;
  pDeviceInfo^.bWindowed := bWindowed;

  // Prepare window for windowed/fullscreen change
  AdjustWindowForChange;

  // Set up the presentation parameters
  d3dpp.Windowed               := pDeviceInfo^.bWindowed;
  d3dpp.MultiSampleType        := pDeviceInfo^.MultiSampleType;
  d3dpp.AutoDepthStencilFormat := pModeInfo^.DepthStencilFormat;
  d3dpp.hDeviceWindow          := hWnd;
  if bWindowed then begin
    d3dpp.BackBufferWidth  := rcWindowClient.right - rcWindowClient.left;
    d3dpp.BackBufferHeight := rcWindowClient.bottom - rcWindowClient.top;
    d3dpp.BackBufferFormat := pAdapterInfo^.d3ddmDesktop.Format;
  end
  else begin
    d3dpp.BackBufferWidth  := pModeInfo^.Width;
    d3dpp.BackBufferHeight := pModeInfo^.Height;
    d3dpp.BackBufferFormat := pModeInfo^.Format;
  end;

  // Resize the 3D device
  if FAILED( Resize3DEnvironment ) then begin
    if bWindowed then begin
      Result := ForceWindowed;
      Exit;
    end
    else begin
      Result := E_FAIL;
      Exit;
    end;
  end;

  // When moving from fullscreen to windowed mode, it is important to
  // adjust the window size after resetting the device rather than
  // beforehand to ensure that you get the window size you want.  For
  // example, when switching from 640x480 fullscreen to windowed with
  // a 1000x600 window on a 1024x768 desktop, it is impossible to set
  // the window size to 1000x600 until after the display mode has
  // changed to 1024x768, because windows cannot be larger than the
  // desktop.
  if bWindowed then begin
    SetWindowPos( hWnd, HWND_NOTOPMOST,
                  rcWindowBounds.left, rcWindowBounds.top,
                  ( rcWindowBounds.right - rcWindowBounds.left ),
                  ( rcWindowBounds.bottom - rcWindowBounds.top ),
                  SWP_SHOWWINDOW );
  end;

  bReady := TRUE;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ForceWindowed()
// Desc: Switch to a windowed mode, even if that means picking a new device
//       and/or adapter
//-----------------------------------------------------------------------------
function TD3DApplication.ForceWindowed : HRESULT;
var
  pAdapterInfo, pAdapterInfoCur : ^TD3DAdapterInfo;
  pDeviceInfo, pDeviceInfoCur : ^TD3DDeviceInfo;
  bFoundDevice : Boolean;
  iAdapter, iDevice : DWORD;
begin
  pAdapterInfoCur := @Adapters[dwAdapter];
  pDeviceInfoCur  := @pAdapterInfoCur^.devices[pAdapterInfoCur^.dwCurrentDevice];
  bFoundDevice   := FALSE;

  if pDeviceInfoCur^.bCanDoWindowed then bFoundDevice := TRUE
  else begin
    // Look for a windowable device on any adapter
    for iAdapter := 0 to dwNumAdapters - 1 do begin
      pAdapterInfo := @Adapters[iAdapter];
      for iDevice := 0 to pAdapterInfo^.dwNumDevices - 1 do begin
        pDeviceInfo := @pAdapterInfo^.devices[iDevice];
        if pDeviceInfo^.bCanDoWindowed then begin
          dwAdapter := iAdapter;
          pDeviceInfoCur := pDeviceInfo;
          pAdapterInfo^.dwCurrentDevice := iDevice;
          bFoundDevice := TRUE;
          break;
        end;
      end;
      if bFoundDevice then break;
    end;
  end;

  if not bFoundDevice then begin
    Result := E_FAIL;
    Exit;
  end;

  pDeviceInfoCur^.bWindowed := TRUE;
  bWindowed := TRUE;

  // Now destroy the current 3D device objects, then reinitialize

  bReady := FALSE;

  // Release all scene objects that will be re-created for the new device
  InvalidateDeviceObjects;
  DeleteDeviceObjects;

  // Release display objects, so a new device can be created
  d3dDevice := nil;
  if d3dDevice <> nil then begin
    Result := DisplayErrorMsg( D3DAPPERR_NONZEROREFCOUNT, MSGERR_APPMUSTEXIT );
    Exit;
  end;

  // Create the new device
  Result := Initialize3DEnvironment;
  if Failed( Result ) then begin
    DisplayErrorMsg( Result, MSGERR_APPMUSTEXIT );
    Exit;
  end;

  bReady := TRUE;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: AdjustWindowForChange()
// Desc: Prepare the window for a possible change between windowed mode and
//       fullscreen mode.  This function is virtual and thus can be overridden
//       to provide different behavior, such as switching to an entirely
//       different window for fullscreen mode (as in the MFC sample apps).
//-----------------------------------------------------------------------------
function TD3DApplication.AdjustWindowForChange : HRESULT;
begin
  if bWindowed then begin
    // Set windowed-mode style
    SetWindowLong( hWnd, GWL_STYLE, dwWindowStyle );
  end
  else begin
    // Set fullscreen-mode style
    SetWindowLong( hWnd, GWL_STYLE, Integer( WS_POPUP or WS_SYSMENU or WS_VISIBLE ) );
  end;
  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Run()
// Desc:
//-----------------------------------------------------------------------------
procedure TD3DApplication.Run( Sender: TObject );
begin
  if bActive and bReady then begin
    if Failed( Render3DEnvironment ) then begin
      bReady := False;
      SendMessage( hWnd, WM_CLOSE, 0, 0 );
    end;
  end;
end;

//-----------------------------------------------------------------------------
// Name: Render3DEnvironment()
// Desc: Draws the scene.
//-----------------------------------------------------------------------------
function TD3DApplication.Render3DEnvironment : HRESULT;
var
  pAdapterInfo : ^TD3DAdapterInfo;
  dwAppTime, dwElapsedAppTime : DWORD;
  mode : TD3DDISPLAYMODE;
  pDeviceInfo : ^TD3DDeviceInfo;
  pModeInfo : ^TD3DModeInfo;
  bpp : Byte;
begin
  // Test the cooperative level to see if it's okay to render
  Result := d3dDevice.TestCooperativeLevel;
  if FAILED( Result ) then begin

    // If the device was lost, do not render until we get it back
    if D3DERR_DEVICELOST = Result then begin
      Result := S_OK;
      Exit;
    end;

    // Check if the device needs to be resized.
    if D3DERR_DEVICENOTRESET = Result then begin
      // If we are windowed, read the desktop mode and use the same format for
      // the back buffer
      if bWindowed then begin
        pAdapterInfo := @Adapters[dwAdapter];
        D3D.GetAdapterDisplayMode( dwAdapter, pAdapterInfo^.d3ddmDesktop );
        d3dpp.BackBufferFormat := pAdapterInfo^.d3ddmDesktop.Format;
      end;

      Result := Resize3DEnvironment;
      if FAILED( Result ) then Exit;
    end;

    Exit;

  end;

  // Get the app's time, in milliseconds. Skip rendering if no time elapsed
  dwAppTime := GetTickCount;
  dwElapsedAppTime := dwAppTime - dwTime;
  if ( 0 = dwElapsedAppTime ) and bFrameMoving then Exit;

  // FrameMove (animate) the scene
  if bFrameMoving or bSingleStep then begin
    // Store the time for the app
    dwTime        := dwAppTime;
    dwElapsedTime := dwElapsedAppTime;

    // Frame move the scene
    Result := FrameMove;
    if Failed( Result ) then Exit;

    bSingleStep := FALSE;
  end;

    // Render the scene as normal
  Result := Render;
  if Failed( Result ) then Exit;

  // Keep track of the frame count
  inc( dwFrames );

  // Update the scene stats once per second
  if( dwAppTime - dwLastTime > 1000 ) then begin
    fFPS := dwFrames / ( dwAppTime - dwLastTime ) * 1000;
    dwLastTime := dwAppTime;
    dwFrames   := 0;

    // Get adapter's current mode so we can report
    // bit depth (back buffer depth may be unknown)
    D3D.GetAdapterDisplayMode( dwAdapter, mode );
    if mode.Format = D3DFMT_X8R8G8B8 then bpp := 32 else bpp := 16;
    strFrameStats := Format( '%8.2f %s (%d %d %d)'
    , [ fFPS, 'fps', d3dsdBackBuffer.Width, d3dsdBackBuffer.Height, bpp ] );

    if bUseDepthBuffer then begin
      pAdapterInfo := @Adapters[dwAdapter];
      pDeviceInfo  := @pAdapterInfo^.devices[pAdapterInfo^.dwCurrentDevice];
      pModeInfo    := @pDeviceInfo^.modes[pDeviceInfo^.dwCurrentMode];

      case pModeInfo^.DepthStencilFormat of
        D3DFMT_D16:     strFrameStats := strFrameStats + ' (D16)';
        D3DFMT_D15S1:   strFrameStats := strFrameStats + ' (D15S1)';
        D3DFMT_D24X8:   strFrameStats := strFrameStats + ' (D24X8)';
        D3DFMT_D24S8:   strFrameStats := strFrameStats + ' (D24S8)';
        D3DFMT_D24X4S4: strFrameStats := strFrameStats + ' (D24X4S4)';
        D3DFMT_D32:     strFrameStats := strFrameStats + ' (D32)';
      end;

    end; // bUseDepthBuffer

  end; // update fps

  // Show the frame on the primary surface.
  d3dDevice.Present( nil, nil, 0, nil );

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Pause()
// Desc: Called in to toggle the pause state of the app.
//-----------------------------------------------------------------------------
procedure TD3DApplication.Pause( bPause : Boolean );
begin
  if bPause then inc( dwAppPausedCount ) else dec( dwAppPausedCount );
  bReady := dwAppPausedCount = 0;
  Timer.Enabled := bReady;
end;

//-----------------------------------------------------------------------------
// Name: Cleanup3DEnvironment()
// Desc: Cleanup scene objects
//-----------------------------------------------------------------------------
procedure TD3DApplication.Cleanup3DEnvironment;
begin
  bActive := FALSE;
  bReady  := FALSE;
  if Timer <> nil then begin
    Timer.Enabled := False;
    Timer.Free;
  end;
  if d3dDevice <> nil then begin
    InvalidateDeviceObjects;
    DeleteDeviceObjects;
    d3dDevice := nil;
    D3D := nil;
  end;
  FinalCleanup;
  end;

function TD3DApplication.DisplayErrorMsg( hr : HRESULT; dwType : DWORD ) : HRESULT;
var
  strMsg : String;
begin
  case hr of
    D3DAPPERR_NODIRECT3D:
      strMsg := 'Could not initialize Direct3D. You may'#13
              + 'want to check that the latest version of'#13
              + 'DirectX is correctly installed on your'#13
              + 'system.  Also make sure that this program'#13
              + 'was compiled with header files that match'#13
              + 'the installed DirectX DLLs.';
    D3DAPPERR_NOCOMPATIBLEDEVICES:
      strMsg := 'Could not find any compatible Direct3D'#13
              + 'devices.';
    D3DAPPERR_NOWINDOWABLEDEVICES:
      strMsg := 'This sample cannot run in a desktop'#13
              + 'window with the current display settings.'#13
              + 'Please change your desktop settings to a'#13
              + '16- or 32-bit display mode and re-run this'#13
              + 'sample.';
    D3DAPPERR_NOHARDWAREDEVICE:
      strMsg := 'No hardware-accelerated Direct3D devices'#13
              + 'were found.';
    D3DAPPERR_HALNOTCOMPATIBLE:
      strMsg := 'This sample requires functionality that is'#13
              + 'not available on your Direct3D hardware'#13
              + 'accelerator.';
    D3DAPPERR_NOWINDOWEDHAL:
      strMsg := 'Your Direct3D hardware accelerator cannot'#13
              + 'render into a window.'#13
              + 'Press F2 while the app is running to see a'#13
              + 'list of available devices and modes.';
    D3DAPPERR_NODESKTOPHAL:
      strMsg := 'Your Direct3D hardware accelerator cannot'#13
              + 'render into a window with the current'#13
              + 'desktop display settings.'#13
              + 'Press F2 while the app is running to see a'#13
              + 'list of available devices and modes.';
    D3DAPPERR_NOHALTHISMODE:
      strMsg := 'This sample requires functionality that is'#13
              + 'not available on your Direct3D hardware'#13
              + 'accelerator with the current desktop display'#13
              + 'settings.'#13
              + 'Press F2 while the app is running to see a'#13
              + 'list of available devices and modes.';
    D3DAPPERR_MEDIANOTFOUND:
      strMsg := 'Could not load required media.';
    D3DAPPERR_RESIZEFAILED:
      strMsg := 'Could not reset the Direct3D device.';
    D3DAPPERR_NONZEROREFCOUNT:
      strMsg := 'A D3D object has a non-zero reference'#13
              + 'count (meaning things were not properly'#13
              + 'cleaned up).';
    E_OUTOFMEMORY:
      strMsg := 'Not enough memory.';
    D3DERR_OUTOFVIDEOMEMORY:
      strMsg := 'Not enough video memory.';
    else
      strMsg := 'Generic application error. Enable'#13
              + 'debug output for detailed information.';
  end;

//  strMsg := StrReplace( strMsg, '\n', chr( 10 ) );

  if MSGERR_APPMUSTEXIT = dwType then begin
    strMsg := strMsg + #13#13'This sample will now exit.';
    MessageBox( 0, PChar( strMsg ), PChar( strWindowTitle ), MB_ICONERROR or MB_OK );

    // Close the window, which shuts down the app
    if hWnd <> 0 then SendMessage( hWnd, WM_CLOSE, 0, 0 );
  end
  else begin
    if MSGWARN_SWITCHEDTOREF = dwType then
      strMsg := strMsg + #13#13'Switching to the reference rasterizer,'#13
                       + 'a software device that implements the entire'#13
                       + 'Direct3D feature set, but runs very slowly.';
    MessageBox( 0, PChar( strMsg ), PChar( strWindowTitle ), MB_ICONWARNING or MB_OK );
  end;

  Result := hr;
end;

end.
