{*********************************************************************
* Copyright (c) 1997-2000 Microsoft Corporation. All rights reserved
*
* File: D3DUtil.h
*
* Desc: Helper functions and typing shortcuts for Direct3D programming.
*
*********************************************************************
* timeline :
*   12/29/2k -> rewritten from c++
*
* by orthkon * orthkon@mail.com * http://mp3.com/orthkon
*
* changes:  added QuickSort routne with callback function
*   J. Rathlev, Nov. 2001 (rathlev@physik.uni-kiel.de)
*
* please note,
* need the included D3DX8, because any functions are new!
*********************************************************************}
unit D3DUtil;

interface

uses Windows, SysUtils, Messages, Math, Direct3D8, D3DX8;

type
  TByteArray = array of byte;
  TCompFunction = function (const arg1,arg2) : boolean of object;

procedure QuickSort (var SList;
                     Count,RSize : Integer;
                     Compare     : TCompFunction);

{-----------------------------------------------------------------------------
/ Name: D3DUtil_InitMaterial()
/ Desc: Initializes a D3DMATERIAL8 structure, setting the diffuse and ambient
/       colors. It does not set emissive or specular colors.
/-----------------------------------------------------------------------------}
procedure D3DUtil_InitMaterial( var mtrl : TD3DMaterial8; r, g, b, a : Single );

//-----------------------------------------------------------------------------
// Name: D3DUtil_InitLight()
// Desc: Initializes a D3DLIGHT structure, setting the light position. The
//       diffuse color is set to white, specular and ambient left as black.
//-----------------------------------------------------------------------------
procedure D3DUtil_InitLight( var light : TD3DLIGHT8; ltType : TD3DLIGHTTYPE; x, y, z : Single );

//-----------------------------------------------------------------------------
// Name: D3DUtil_CreateTexture()
// Desc: Helper function to create a texture. It checks the root path first,
//       then tries the DXSDK media path (as specified in the system registry).
//-----------------------------------------------------------------------------
function D3DUtil_CreateTexture( d3dDevice : IDirect3DDevice8; strTexture : PChar; out ppTexture : IDirect3DTexture8; d3dFormat : TD3DFORMAT ) : HRESULT;

//-----------------------------------------------------------------------------
// Name: D3DUtil_SetColorKey()
// Desc: Changes all texels matching the colorkey to transparent, black.
//-----------------------------------------------------------------------------
function D3DUtil_SetColorKey( var pTexture : IDirect3DTexture8; dwColorKey : DWORD ) : HRESULT;

//-----------------------------------------------------------------------------
// Name: D3DUtil_CreateVertexShader()
// Desc: Assembles and creates a file-based vertex shader
//-----------------------------------------------------------------------------
function D3DUtil_CreateVertexShader( d3dDevice : IDirect3DDevice8; strFilename : PChar; pdwVertexDecl : DWORD; pdwVertexShader : DWORD ) : HRESULT;

//-----------------------------------------------------------------------------
// Name: D3DUtil_GetCubeMapViewMatrix()
// Desc: Returns a view matrix for rendering to a face of a cubemap.
//-----------------------------------------------------------------------------
function D3DUtil_GetCubeMapViewMatrix( dwFace : TD3DCubemapFaces) : TD3DXMATRIX;

//-----------------------------------------------------------------------------
// Name: D3DUtil_GetRotationFromCursor()
// Desc: Returns a quaternion for the rotation implied by the window's cursor
//       position.
//-----------------------------------------------------------------------------
function D3DUtil_GetRotationFromCursor( hWnd : Integer; fTrackBallRadius : Single ) : TD3DXQUATERNION;

//-----------------------------------------------------------------------------
// Name: D3DUtil_SetDeviceCursor
// Desc: Builds and sets a cursor for the D3D device based on hCursor.
//-----------------------------------------------------------------------------
function D3DUtil_SetDeviceCursor( d3dDevice : IDirect3DDevice8; hCursor : HCURSOR ) : HRESULT;


//-----------------------------------------------------------------------------
// Helper macros for pixel shader instructions
//-----------------------------------------------------------------------------

// Parameter writemasks
const D3DPSP_WRITEMASK_B = D3DSP_WRITEMASK_0;
const D3DPSP_WRITEMASK_G = D3DSP_WRITEMASK_1;
const D3DPSP_WRITEMASK_R = D3DSP_WRITEMASK_2;
const D3DPSP_WRITEMASK_A = D3DSP_WRITEMASK_3;
const D3DPSP_WRITEMASK_C = D3DPSP_WRITEMASK_B or D3DPSP_WRITEMASK_G or D3DPSP_WRITEMASK_R;
const D3DPSP_WRITEMASK_ALL = D3DSP_WRITEMASK_0 or D3DSP_WRITEMASK_1 or D3DSP_WRITEMASK_2 or D3DSP_WRITEMASK_3;
const D3DPSP_WRITEMASK_10 = D3DSP_WRITEMASK_0 or D3DSP_WRITEMASK_1;
const D3DPSP_WRITEMASK_32 = D3DSP_WRITEMASK_2 or D3DSP_WRITEMASK_3;

// Source and destination parameter token
function D3DPS_REGNUM_MASK( _Num : Integer ) : Integer;
function D3DPS_DST( _Num : Integer ) : Integer;
function D3DPS_SRC_TEMP( _Num : Integer ) : Integer;
function D3DPS_SRC_INPUT( _Num : Integer ) : Integer;
function D3DPS_SRC_CONST( _Num : Integer ) : Integer;
function D3DPS_SRC_TEXTURE( _Num : Integer ) : Integer;
function D3DVS_SRC_ADDR( _Num : Integer ) : Integer;
function D3DVS_SRC_RASTOUT( _Num : Integer ) : Integer;
function D3DVS_SRC_ATTROUT( _Num : Integer ) : Integer;
function D3DVS_SRC_TEXCRDOUT( _Num : Integer ) : Integer;

// Temp destination registers -> D3DPS_DST( 0 to 7 )
const D3DS_DR0 = ( ( ( 1 shl 31 ) or ( ( 0 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSPR_TEMP or D3DPSP_WRITEMASK_ALL;
const D3DS_DR1 = ( ( ( 1 shl 31 ) or ( ( 1 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSPR_TEMP or D3DPSP_WRITEMASK_ALL;
const D3DS_DR2 = ( ( ( 1 shl 31 ) or ( ( 2 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSPR_TEMP or D3DPSP_WRITEMASK_ALL;
const D3DS_DR3 = ( ( ( 1 shl 31 ) or ( ( 3 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSPR_TEMP or D3DPSP_WRITEMASK_ALL;
const D3DS_DR4 = ( ( ( 1 shl 31 ) or ( ( 4 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSPR_TEMP or D3DPSP_WRITEMASK_ALL;
const D3DS_DR5 = ( ( ( 1 shl 31 ) or ( ( 5 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSPR_TEMP or D3DPSP_WRITEMASK_ALL;
const D3DS_DR6 = ( ( ( 1 shl 31 ) or ( ( 6 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSPR_TEMP or D3DPSP_WRITEMASK_ALL;
const D3DS_DR7 = ( ( ( 1 shl 31 ) or ( ( 7 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSPR_TEMP or D3DPSP_WRITEMASK_ALL;

// Temp source registers -> D3DPS_SRC_TEMP( 0 to 7 )
const D3DS_SR0 = ( ( ( 1 shl 31 ) or ( ( 0 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEMP;
const D3DS_SR1 = ( ( ( 1 shl 31 ) or ( ( 1 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEMP;
const D3DS_SR2 = ( ( ( 1 shl 31 ) or ( ( 2 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEMP;
const D3DS_SR3 = ( ( ( 1 shl 31 ) or ( ( 3 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEMP;
const D3DS_SR4 = ( ( ( 1 shl 31 ) or ( ( 4 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEMP;
const D3DS_SR5 = ( ( ( 1 shl 31 ) or ( ( 5 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEMP;
const D3DS_SR6 = ( ( ( 1 shl 31 ) or ( ( 6 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEMP;
const D3DS_SR7 = ( ( ( 1 shl 31 ) or ( ( 7 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEMP;

// Texture parameters -> D3DPS_SRC_TEXTURE( 0 to 7 )
const D3DS_T0 = ( ( ( 1 shl 31 ) or ( ( 0 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEXTURE;
const D3DS_T1 = ( ( ( 1 shl 31 ) or ( ( 1 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEXTURE;
const D3DS_T2 = ( ( ( 1 shl 31 ) or ( ( 2 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEXTURE;
const D3DS_T3 = ( ( ( 1 shl 31 ) or ( ( 3 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEXTURE;
const D3DS_T4 = ( ( ( 1 shl 31 ) or ( ( 4 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEXTURE;
const D3DS_T5 = ( ( ( 1 shl 31 ) or ( ( 5 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEXTURE;
const D3DS_T6 = ( ( ( 1 shl 31 ) or ( ( 6 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEXTURE;
const D3DS_T7 = ( ( ( 1 shl 31 ) or ( ( 7 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEXTURE;

// Constant (factor) source parameters -> D3DPS_SRC_CONST( 0 to 7 )
const D3DS_C0 = ( ( ( 1 shl 31 ) or ( ( 0 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_CONST;
const D3DS_C1 = ( ( ( 1 shl 31 ) or ( ( 1 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_CONST;
const D3DS_C2 = ( ( ( 1 shl 31 ) or ( ( 2 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_CONST;
const D3DS_C3 = ( ( ( 1 shl 31 ) or ( ( 3 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_CONST;
const D3DS_C4 = ( ( ( 1 shl 31 ) or ( ( 4 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_CONST;
const D3DS_C5 = ( ( ( 1 shl 31 ) or ( ( 5 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_CONST;
const D3DS_C6 = ( ( ( 1 shl 31 ) or ( ( 6 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_CONST;
const D3DS_C7 = ( ( ( 1 shl 31 ) or ( ( 7 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_CONST;

// Iterated source parameters ( 0 = Diffuse, 1 = specular )
const D3DS_V0 = ( ( ( 1 shl 31 ) or ( ( 0 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_INPUT;
const D3DS_V1 = ( ( ( 1 shl 31 ) or ( ( 1 ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_INPUT;

type
  TD3DArcBall = class
	  m_iWidth : Integer;  // ArcBall's window width
	  m_iHeight : Integer; // ArcBall's window height
	  m_fRadius : Single;  // ArcBall's radius in screen coords
	  m_fRadius2 : Single; // ArcBall's radius in screen coords

	  m_qDown : TD3DXQUATERNION;           // Quaternion before button down
	  m_qNow : TD3DXQUATERNION;            // Composite quaternion for current drag
	  m_matRotation : TD3DXMATRIX;         // Matrix for arcball's orientation
	  m_matRotationDelta : TD3DXMATRIX;    // Matrix for arcball's orientation
	  m_matTranslation : TD3DXMATRIX;      // Matrix for arcball's position
	  m_matTranslationDelta : TD3DXMATRIX; // Matrix for arcball's position
	  m_bDrag : Boolean;                   // Whether user is dragging arcball

    iMouseX, iMouseY : Integer; // Current mouse position
    iCurMouseX, iCurMouseY : Integer; // static
    m_vDown : TD3DXVECTOR3;    // Button down quaternion
    m_vCur : TD3DXVECTOR3;     // Current quaternion

  	function ScreenToVector( sx, sy : Integer ) : TD3DXVECTOR3;

  public
    constructor Create; overload;
    function HandleMouseMessages( hWnd : HWND; uMsg, wParam : WORD; lParam : DWORD ) : Boolean;

    function GetRotationMatrix : TD3DXMATRIX;
    function GetRotationDeltaMatrix : TD3DXMATRIX;
    function GetTranslationMatrix : TD3DXMATRIX;
    function GetTranslationDeltaMatrix : TD3DXMATRIX;
    function IsBeingDragged : Boolean;

    procedure SetRadius( fRadius : Single );
    procedure SetWindow( iWidth, iHeight : Integer; fRadius : Single );
  end;

  TD3DCamera = class
  	m_vEyePt       : TD3DXVECTOR3; // Attributes for view matrix
	  m_vLookatPt    : TD3DXVECTOR3;
	  m_vUpVec       : TD3DXVECTOR3;

	  m_vView        : TD3DXVECTOR3;
	  m_vCross       : TD3DXVECTOR3;

	  m_matView      : TD3DXMATRIX;
	  m_matBillboard : TD3DXMATRIX; // Special matrix for billboarding effects

	  m_fFOV         : Single;      // Attributes for projection matrix
	  m_fAspect      : Single;
	  m_fNearPlane   : Single;
	  m_fFarPlane    : Single;
	  m_matProj      : TD3DXMATRIX;

  public
    constructor Create; overload;

  	// Access functions
	  function GetEyePt : TD3DXVECTOR3;
	  function GetLookatPt : TD3DXVECTOR3;
	  function GetUpVec : TD3DXVECTOR3;
	  function GetViewDir : TD3DXVECTOR3;
	  function GetCross : TD3DXVECTOR3;

	  function GetViewMatrix : TD3DXMATRIX;
	  function GetBillboardMatrix : TD3DXMATRIX;
	  function GetProjMatrix : TD3DXMATRIX;

    procedure SetViewParams( vEyePt, vLookatPt, vUpVec : TD3DXVECTOR3 );
    procedure SetProjParams( fFOV, fAspect, fNearPlane, fFarPlane : Single );
  end;

{ added functions for vector arithmetic
  taken from: Neobrothers Tutorial: www.neobrothers.de: Ne8_d3d.pas
}

function MakeVector(x, y, z: single): TD3DVector;
function VectorNormalize(v1: TD3DXVector3): TD3DXVector3;
function VectorCrossproduct(v1, v2: TD3DXVector3): TD3DXVector3;
function VectorDotproduct(v1, v2: TD3DXVector3): single;
function VectorAdd(v1, v2: TD3DXVector3): TD3DXVector3;
function VectorSub(v1, v2: TD3DXVector3): TD3DXVector3;
function VectorMul(v1, v2: TD3DXVector3): TD3DXVector3;
function VectorMuls(v1: TD3DXVector3; s: single): TD3DXVector3;
function VectorDiv(v1, v2: TD3DXVector3): TD3DXVector3;
function VectorDivs(v1: TD3DXVector3; s: single): TD3DXVector3;
function VectorMagnitude(v1: TD3DXVector3): single;


implementation

const D3DX_DEFAULT : DWORD = $ffffffff;

function D3DPS_REGNUM_MASK( _Num : Integer ) : Integer; begin Result := ( ( 1 shl 31 ) or ( ( _Num ) and D3DSP_REGNUM_MASK ) ); end;
function D3DPS_DST( _Num : Integer ) : Integer; begin Result := ( ( ( 1 shl 31 ) or ( ( _Num ) and D3DSP_REGNUM_MASK ) ) ) or D3DSPR_TEMP or D3DPSP_WRITEMASK_ALL; end;
function D3DPS_SRC_TEMP( _Num : Integer ) : Integer; begin Result := ( ( ( 1 shl 31 ) or ( ( _Num ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEMP; end;
function D3DPS_SRC_INPUT( _Num : Integer ) : Integer; begin Result := ( ( ( 1 shl 31 ) or ( ( _Num ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_INPUT; end;
function D3DPS_SRC_CONST( _Num : Integer ) : Integer; begin Result := ( ( ( 1 shl 31 ) or ( ( _Num ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_CONST; end;
function D3DPS_SRC_TEXTURE( _Num : Integer ) : Integer; begin Result := ( ( ( 1 shl 31 ) or ( ( _Num ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEXTURE; end;
function D3DVS_SRC_ADDR( _Num : Integer ) : Integer; begin Result := ( ( ( 1 shl 31 ) or ( ( _Num ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_ADDR; end;
function D3DVS_SRC_RASTOUT( _Num : Integer ) : Integer; begin Result := ( ( ( 1 shl 31 ) or ( ( _Num ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_RASTOUT; end;
function D3DVS_SRC_ATTROUT( _Num : Integer ) : Integer; begin Result := ( ( ( 1 shl 31 ) or ( ( _Num ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_ATTROUT; end;
function D3DVS_SRC_TEXCRDOUT( _Num : Integer ) : Integer; begin Result := ( ( ( 1 shl 31 ) or ( ( _Num ) and D3DSP_REGNUM_MASK ) ) ) or D3DSP_NOSWIZZLE or D3DSPR_TEXCRDOUT; end;

procedure D3DUtil_InitMaterial( var mtrl : TD3DMaterial8; r, g, b, a : Single );
begin
  ZeroMemory( @mtrl , SizeOf( mtrl ) );
  mtrl.Diffuse.r := r;
  mtrl.Diffuse.g := g;
  mtrl.Diffuse.b := b;
  mtrl.Diffuse.a := a;
  mtrl.Ambient := mtrl.Diffuse;
end;

procedure D3DUtil_InitLight( var light : TD3DLIGHT8; ltType : TD3DLIGHTTYPE; x, y, z : Single );
var
  t3dv : TD3DVector;
begin
  ZeroMemory( @light , SizeOf( light ) );
  light._Type := ltType;
  light.Diffuse.r := 1;
  light.Diffuse.g := 1;
  light.Diffuse.b := 1;
  t3dv := D3DXVector3( x, y, z );
  D3DXVec3Normalize( light.Direction, t3dv );
  light.Position.x := x;
  light.Position.y := y;
  light.Position.z := z;
  light.Range := 1000;
end;

function D3DUtil_CreateTexture( d3dDevice : IDirect3DDevice8; strTexture : PChar; out ppTexture : IDirect3DTexture8; d3dFormat : TD3DFORMAT ) : HRESULT;
begin
  Result := D3DXCreateTextureFromFileEx( d3dDevice, strTexture,
    D3DX_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, d3dFormat,
    D3DPOOL_MANAGED, D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR,
    D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR, 0, nil, nil, ppTexture );
end;

function D3DUtil_SetColorKey( var pTexture : IDirect3DTexture8; dwColorKey : DWORD ) : HRESULT;
var
  r, g, b, y, x : DWORD;
  d3dsd : TD3DSURFACE_DESC;
  d3dlr : TD3DLOCKED_RECT;
  pwbits : ^WORD;
  pdwbits : ^DWORD;
begin
  Result := E_FAIL;

  // Get colorkey's red, green, and blue components
  r := $ff and ( dwColorKey shr 16 );
  g := $ff and ( dwColorKey shr 8 );
  b := $ff and dwColorKey;

  // Put the colorkey in the texture's native format
  pTexture.GetLevelDesc( 0, d3dsd );
  if d3dsd.Format = D3DFMT_A4R4G4B4 then
    dwColorKey := $f000 + ( ( r shr 4 ) shl 8 ) + ( ( g shr 4 ) shl 4 ) + ( b shr 4 )
  else if d3dsd.Format = D3DFMT_A1R5G5B5 then
    dwColorKey := $8000 + ( ( r shr 3 ) shl 10 ) + ( ( g shr 3 ) shl 5 ) + ( b shr 3 )
  else if d3dsd.Format <> D3DFMT_A8R8G8B8 then Exit;

  // Lock the texture
  if Failed( pTexture.LockRect( 0, d3dlr, nil, 0 ) ) then Exit;

  // load the pointer as word and dword values
  pwbits := d3dlr.pBits;
  pdwbits := d3dlr.pBits;

  // Scan through each pixel, looking for the colorkey to replace
  for y := 0 to d3dsd.Height - 1 do
  for x := 0 to d3dsd.Width - 1 do begin
    if d3dsd.Format = D3DFMT_A8R8G8B8 then begin
      // Handle 32-bit formats
      if pdwbits^ = dwColorKey then pdwbits^ := $00000000;
      inc( pdwbits );
    end else begin
      // Handle 16-bit formats
      if pwbits^ = dwColorKey then pwbits^ := $0000;
      inc( pwbits );
    end;
  end;

  // Unlock the texture and return OK.
  pTexture.UnlockRect( 0 );
  Result := S_OK;
end;

function D3DUtil_CreateVertexShader( d3dDevice : IDirect3DDevice8; strFilename : PChar; pdwVertexDecl : DWORD; pdwVertexShader : DWORD ) : HRESULT;
var
 pCode : ID3DXBUFFER;
 bp : PLongWORD;
begin
  pCode := nil;
  // Assemble the vertex shader file
  Result := D3DXAssembleShaderFromFile( strFilename, 0, nil, pCode, nil );
  if Failed( Result ) then Exit;

  // Create the vertex shader
  bp := pCode.GetBufferPointer();
  Result := d3dDevice.CreateVertexShader( pdwVertexDecl, bp, pdwVertexShader, 0 );
  pCode._Release;
end;

function D3DUtil_GetCubeMapViewMatrix( dwFace : TD3DCubemapFaces) : TD3DXMATRIX;
var
  vEyePt, vLookDir, vUpDir : TD3DXVECTOR3;
  matView : TD3DXMATRIX ;
begin
  vEyePt :=  D3DXVECTOR3( 0, 0, 0 );
  if dwFace = D3DCUBEMAP_FACE_POSITIVE_X then begin
    vLookDir := D3DXVECTOR3( 1.0, 0.0, 0.0 );
    vUpDir   := D3DXVECTOR3( 0.0, 1.0, 0.0 );
  end;
  if dwFace = D3DCUBEMAP_FACE_NEGATIVE_X then begin
    vLookDir := D3DXVECTOR3(-1.0, 0.0, 0.0 );
    vUpDir   := D3DXVECTOR3( 0.0, 1.0, 0.0 );
  end;
  if dwFace = D3DCUBEMAP_FACE_POSITIVE_Y then begin
    vLookDir := D3DXVECTOR3( 0.0, 1.0, 0.0 );
    vUpDir   := D3DXVECTOR3( 0.0, 0.0,-1.0 );
  end;
  if dwFace = D3DCUBEMAP_FACE_NEGATIVE_Y then begin
    vLookDir := D3DXVECTOR3( 0.0,-1.0, 0.0 );
    vUpDir   := D3DXVECTOR3( 0.0, 0.0, 1.0 );
  end;
  if dwFace = D3DCUBEMAP_FACE_POSITIVE_Z then begin
    vLookDir := D3DXVECTOR3( 0.0, 0.0, 1.0 );
    vUpDir   := D3DXVECTOR3( 0.0, 1.0, 0.0 );
  end;
  if dwFace = D3DCUBEMAP_FACE_NEGATIVE_Z then begin
    vLookDir := D3DXVECTOR3( 0.0, 0.0,-1.0 );
    vUpDir   := D3DXVECTOR3( 0.0, 1.0, 0.0 );
  end;

  // Set the view transform for this cubemap surface
  D3DXMatrixLookAtLH( matView, vEyePt, vLookDir, vUpDir );
  Result := matView;
end;

function D3DUtil_GetRotationFromCursor( hWnd : Integer; fTrackBallRadius : Single ) : TD3DXQUATERNION;
var
  pt : TPOINT;
  rc : TRECT;
  fAngle, t, d2, sx, sy, sz : Single;
  p1, p2, p3, vAxis : TD3DXVECTOR3;
begin
  GetCursorPos( pt );
  GetClientRect( hWnd, rc );
  ScreenToClient( hWnd, pt );
  sx := ( ( ( 2.0 * pt.x ) / ( rc.right - rc.left ) ) - 1 );
  sy := ( ( ( 2.0 * pt.y ) / ( rc.bottom - rc.top ) ) - 1 );

  if ( sx = 0.0 ) and ( sy = 0.0 ) then begin
    Result := D3DXQUATERNION( 0.0, 0.0, 0.0, 1.0 );
    Exit;
  end;

  //d1 := 0;
  d2 := sqrt( sqr( sx ) + sqr( sy ) );

  if d2 < fTrackBallRadius * 0.70710678118654752440 then // Inside sphere
    sz := sqrt( sqr( fTrackBallRadius ) - sqr( d2 ) )
  else                                                 // On hyperbola
    sz := sqr( fTrackBallRadius ) / ( 2 * d2 );

    // Get two points on trackball's sphere
  p1 := D3DXVECTOR3( sx, sy, sz );
  p2 := D3DXVECTOR3( 0, 0, fTrackBallRadius );

  // Get axis of rotation, which is cross product of p1 and p2
  D3DXVec3Cross( vAxis, p1, p2 );

  // Calculate angle for the rotation about that axis
  D3DXVec3Subtract( p3, p2, p1 );
  t := D3DXVec3Length( p3 ) / ( 2 * fTrackBallRadius );
  if t > 1 then t := 1;
  if t < -1 then t := -1;
  fAngle := 2 * ArcSin( t );

    // Convert axis to quaternion
  D3DXQuaternionRotationAxis( Result, vAxis, fAngle );
end;

function D3DUtil_SetDeviceCursor( d3dDevice : IDirect3DDevice8; hCursor : HCURSOR ) : HRESULT;
label EndP;
var
  iconinfo : Windows.ICONINFO;
  bBWCursor : Boolean;
  pCursorBitmap : IDIRECT3DSURFACE8;
  hdcColor, hdcMask, hdcScreen : Windows.HDC;
  bm : Windows.BITMAP;
  dwWidth, dwHeightSrc, dwHeightDest : DWORD;
  crColor, crMask : Cardinal;
  x, y : WORD;
  bmi : BITMAPINFO;
  pcrArrayColor, pcrArrayMask : array of Cardinal;
  pdwbits : ^DWORD;
  hgdiobjOld : Windows.HGDIOBJ;
  lr : TD3DLOCKED_RECT;
begin
  Result := E_FAIL;
  pCursorBitmap := nil;
  hdcColor := 0;
  hdcMask := 0;
  hdcScreen := 0;

  ZeroMemory( @iconinfo, sizeof(iconinfo) );
  if not GetIconInfo( hCursor, iconinfo ) then goto EndP;

  if 0 = GetObject( iconinfo.hbmMask, sizeof( BITMAP ),  @bm )
    then goto EndP;
  dwWidth := bm.bmWidth;
  dwHeightSrc := bm.bmHeight;

  if iconinfo.hbmColor = 0 then begin
    bBWCursor := True;
    dwHeightDest := dwHeightSrc div 2;
  end else begin
    bBWCursor := FALSE;
    dwHeightDest := dwHeightSrc;
  end;

  // Create a surface for the fullscreen cursor
  Result := d3dDevice.CreateImageSurface( dwWidth, dwHeightDest
    , D3DFMT_A8R8G8B8, pCursorBitmap );
  if Failed( Result ) then goto EndP;

  try SetLength( pcrArrayMask, dwWidth * dwHeightSrc );
    except Result := E_FAIL; end;
  if Failed( Result ) then goto EndP;

  ZeroMemory( @bmi, sizeof( bmi ) );
  with bmi.bmiHeader do begin
    biSize := sizeof( bmi.bmiHeader );
    biWidth := dwWidth;
    biHeight := dwHeightSrc;
    biPlanes := 1;
    biBitCount := 32;
    biCompression := BI_RGB;
  end;

  hdcScreen := GetDC( 0 );
  hdcMask := CreateCompatibleDC( hdcScreen );
  if hdcMask = 0 then begin
    Result := E_FAIL;
    goto EndP;
  end;

  hgdiobjOld := SelectObject( hdcMask, iconinfo.hbmMask );
  GetDIBits( hdcMask, iconinfo.hbmMask, 0, dwHeightSrc
    , pcrArrayMask, bmi, DIB_RGB_COLORS );
  SelectObject( hdcMask, hgdiobjOld );

  if not bBWCursor then begin
    try SetLength( pcrArrayColor, dwWidth * dwHeightDest );
      except Result := E_FAIL; end;
    if Failed( Result ) then goto EndP;

    hdcColor := CreateCompatibleDC( GetDC( 0 ) );
    if hdcColor = 0 then begin
      Result := E_FAIL;
      goto EndP;
    end;
    SelectObject( hdcColor, iconinfo.hbmColor );
    GetDIBits( hdcColor, iconinfo.hbmColor, 0, dwHeightDest
      , pcrArrayColor, bmi, DIB_RGB_COLORS );
  end;

  // Transfer cursor image into the surface
  pCursorBitmap.LockRect( lr, nil, 0 );
  pdwbits := lr.pBits;
  for y := 0 to dwHeightDest - 1 do
  for x := 0 to dwWidth - 1 do begin
    if bBWCursor then begin
      crColor := pcrArrayMask[dwWidth*(dwHeightDest-1-y)+x];
      crMask := pcrArrayMask[dwWidth*(dwHeightSrc-1-y)+x];
    end else begin
      crColor := pcrArrayColor[dwWidth*(dwHeightDest-1-y)+ x];
      crMask := pcrArrayMask[dwWidth*(dwHeightDest-1-y)+x];
    end;
    if crMask = 0 then
      pdwbits^ := $ff000000 or crColor
    else
      pdwbits^ := 0;
    inc( pdwbits );
  end;

  pCursorBitmap.UnlockRect;

  // Set the device cursor
  Result := d3dDevice.SetCursorProperties( iconinfo.xHotspot
    , iconinfo.yHotspot, pCursorBitmap );
  if Failed( Result ) then goto EndP;

  Result := S_OK;
EndP:
  if iconinfo.hbmMask <> 0 then DeleteObject( iconinfo.hbmMask );
  if iconinfo.hbmColor <> 0 then DeleteObject( iconinfo.hbmColor );
  if hdcScreen <> 0 then ReleaseDC( 0, hdcScreen );
  if hdcColor <> 0 then DeleteDC( hdcColor );
  if hdcMask <> 0 then DeleteDC( hdcMask );
  pcrArrayColor := nil;
  pcrArrayMask  := nil;
  pCursorBitmap := nil;
end;

constructor TD3DArcBall.Create;
begin
  D3DXQuaternionIdentity( m_qDown );
  D3DXQuaternionIdentity( m_qNow );
  D3DXMatrixIdentity( m_matRotation );
  D3DXMatrixIdentity( m_matRotationDelta );
  D3DXMatrixIdentity( m_matTranslation );
  D3DXMatrixIdentity( m_matTranslationDelta );
  m_bDrag := FALSE;
end;

procedure TD3DArcBall.SetWindow( iWidth, iHeight : Integer; fRadius : Single );
begin
    // Set ArcBall info
  m_iWidth  := iWidth;
  m_iHeight := iHeight;
  m_fRadius := fRadius;
end;

function TD3DArcBall.ScreenToVector( sx, sy : Integer ) : TD3DXVECTOR3;
var
  x, y, z, mag, scale : Single;
begin
  // Scale to screen
  x   :=  ( sx - m_iWidth / 2 )  / ( m_fRadius * m_iWidth / 2 );
  y   := -( sy - m_iHeight / 2 ) / ( m_fRadius * m_iHeight / 2 );
  z   := 0;
  mag := sqr( x ) + sqr( y );

  if mag > 1 then begin
    scale := 1 / sqrt( mag );
    x := x * scale;
    y := y * scale;
  end else z := sqrt( 1 - mag );

  // Return vector
  Result := D3DXVECTOR3( x, y, z );
end;

procedure TD3DArcBall.SetRadius( fRadius : Single );
begin
  m_fRadius2 := fRadius;
end;

function TD3DArcBall.HandleMouseMessages( hWnd : HWND; uMsg, wParam : WORD; lParam : DWORD ) : Boolean;
var
  vPart, vCur : TD3DXVECTOR3;
  vt,qConj : TD3DXQUATERNION;
  fDeltaX, fDeltaY : Single;

begin
  Result := true;

  iMouseY := LOWORD( lParam );
  iMouseY := HIWORD( lParam );

  case uMsg of
    WM_RBUTTONDOWN,
    WM_MBUTTONDOWN: begin
      // Store off the position of the cursor when the button is pressed
      iCurMouseX := iMouseX;
      iCurMouseY := iMouseY;
      Exit;
    end;
    WM_LBUTTONDOWN: begin
      // Start drag mode
      m_bDrag := TRUE;
      m_vDown := ScreenToVector( iMouseX, iMouseY );
      Exit;
    end;
    WM_LBUTTONUP: begin
      // End drag mode
      m_bDrag := FALSE;
      m_qDown := m_qNow;
      Exit;
    end;
    WM_MOUSEMOVE: begin
      // Drag object
      if ( MK_LBUTTON and wParam ) = MK_LBUTTON then begin
        if m_bDrag then begin
          vCur := ScreenToVector( iMouseX, iMouseY );
          D3DXVec3Cross( vPart, m_vDown, vCur );
          vt:=D3DXQUATERNION(vPart.x, vPart.y, vPart.z, D3DXVec3Dot(m_vDown, vCur));
          m_qNow := D3DXQuaternionMul(m_qDown,vt);
        end;
        D3DXQuaternionConjugate( qConj, m_qNow );

        D3DXMatrixRotationQuaternion( m_matRotationDelta, qConj );
        D3DXMatrixTranspose( m_matRotationDelta, m_matRotationDelta );
        D3DXMatrixMultiply( m_matRotation, m_matRotation, m_matRotationDelta );

        D3DXQuaternionIdentity( m_qDown );
        D3DXQuaternionIdentity( m_qNow );
        m_vDown := ScreenToVector( iMouseX, iMouseY );
        m_bDrag := TRUE;
      end
      else if ( ( MK_RBUTTON and wParam ) = MK_RBUTTON )
      or ( ( MK_MBUTTON and wParam ) = MK_MBUTTON ) then begin
        // Normalize based on size of window and bounding sphere radius
        fDeltaX := ( iCurMouseX-iMouseX ) * m_fRadius2 / m_iWidth;
        fDeltaY := ( iCurMouseY-iMouseY ) * m_fRadius2 / m_iHeight;

        if ( wParam and MK_RBUTTON ) = MK_RBUTTON then begin
          D3DXMatrixTranslation( m_matTranslationDelta, -2 * fDeltaX, 2 * fDeltaY, 0.0 );
          D3DXMatrixMultiply( m_matTranslation, m_matTranslation, m_matTranslationDelta );
        end else begin  // wParam & MK_MBUTTON
          D3DXMatrixTranslation( m_matTranslationDelta, 0.0, 0.0, 5 * fDeltaY );
          D3DXMatrixMultiply( m_matTranslation, m_matTranslation, m_matTranslationDelta );
        end;
        // Store mouse coordinate
        iCurMouseX := iMouseX;
        iCurMouseY := iMouseY;
      end;
      Exit;
    end;
  end;
  Result := FALSE;
end;

function TD3DArcBall.GetRotationMatrix : TD3DXMATRIX;
begin
  Result := m_matRotation;
end;

function TD3DArcBall.GetRotationDeltaMatrix : TD3DXMATRIX;
begin
  Result := m_matRotationDelta;
end;

function TD3DArcBall.GetTranslationMatrix : TD3DXMATRIX;
begin
  Result := m_matTranslation;
end;

function TD3DArcBall.GetTranslationDeltaMatrix : TD3DXMATRIX;
begin
  Result := m_matTranslationDelta;
end;

function TD3DArcBall.IsBeingDragged : Boolean;
begin
  Result := m_bDrag;
end;

constructor TD3DCamera.Create;
begin
  // Set attributes for the view matrix
  SetViewParams
    ( D3DXVECTOR3( 0.0, 0.0, 0.0 )
    , D3DXVECTOR3( 0.0, 0.0, 1.0 )
    , D3DXVECTOR3( 0.0, 1.0, 0.0 )
  );

  // Set attributes for the projection matrix
  SetProjParams( D3DX_PI / 4, 1.0, 1.0, 1000.0 );
end;

procedure TD3DCamera.SetViewParams( vEyePt, vLookatPt, vUpVec : TD3DXVECTOR3 );
var
  tv1 : TD3DXVECTOR3;
begin
  // Set attributes for the view matrix
  m_vEyePt    := vEyePt;
  m_vLookatPt := vLookatPt;
  m_vUpVec    := vUpVec;
  D3DXVec3Subtract( tv1, m_vLookatPt, m_vEyePt );
  D3DXVec3Normalize( m_vView, tv1 );
  D3DXVec3Cross( m_vCross, m_vView, m_vUpVec );

  D3DXMatrixLookAtLH( m_matView, m_vEyePt, m_vLookatPt, m_vUpVec );
  D3DXMatrixInverse( m_matBillboard, nil, m_matView );
  m_matBillboard._41 := 0.0;
  m_matBillboard._42 := 0.0;
  m_matBillboard._43 := 0.0;
end;

procedure TD3DCamera.SetProjParams( fFOV, fAspect, fNearPlane, fFarPlane : Single );
begin
  // Set attributes for the projection matrix
  m_fFOV        := fFOV;
  m_fAspect     := fAspect;
  m_fNearPlane  := fNearPlane;
  m_fFarPlane   := fFarPlane;

  D3DXMatrixPerspectiveFovLH( m_matProj, fFOV, fAspect, fNearPlane, fFarPlane );
end;

function TD3DCamera.GetEyePt : TD3DXVECTOR3; begin Result := m_vEyePt; end;
function TD3DCamera.GetLookatPt : TD3DXVECTOR3; begin Result := m_vLookatPt; end;
function TD3DCamera.GetUpVec : TD3DXVECTOR3; begin Result := m_vUpVec; end;
function TD3DCamera.GetViewDir : TD3DXVECTOR3; begin Result := m_vView; end;
function TD3DCamera.GetCross : TD3DXVECTOR3; begin Result := m_vCross; end;
function TD3DCamera.GetViewMatrix : TD3DXMATRIX; begin Result := m_matView; end;
function TD3DCamera.GetBillboardMatrix : TD3DXMATRIX; begin Result := m_matBillboard; end;
function TD3DCamera.GetProjMatrix : TD3DXMATRIX; begin Result := m_matProj; end;

//##############################################################################
//  additional functions
//  from: Neobrothers Tutorial: www.neobrothers.de: Ne8_d3d.pas
//##############################################################################
function MakeVector(x, y, z: single): TD3DVector;
begin
  with result do begin
    x := x; y := y; z := z;
    end;
  end;

function VectorNormalize(v1: TD3DXVector3): TD3DXVector3;
var
  length : single;
begin
  length:=sqrt(sqr(v1.x) + sqr(v1.y) + sqr(v1.z));
  with result do begin
    x := v1.x / length; y := v1.y / length; z := v1.z / length;
    end;
  end;

function VectorCrossproduct(v1, v2: TD3DXVector3): TD3DXVector3;
begin
  with result do begin
    x := (v1.y*v2.z) - (v1.z*v2.y);
    y := (v1.z*v2.x) - (v1.x*v2.z);
    z := (v1.x*v2.y) - (v1.y*v2.x);
    end;
  end;

function VectorDotproduct(v1, v2: TD3DXVector3): single;
begin
  result := (v1.x * v2.x) + (v1.y * v2.y) + (v1.z * v2.z);
  end;

function VectorAdd(v1, v2: TD3DXVector3): TD3DXVector3;
begin
  with result do begin
    x := v1.x + v2.x; y := v1.y + v2.y; z := v1.z + v2.z;
    end;
  end;

function VectorSub(v1, v2: TD3DXVector3): TD3DXVector3;
begin
  with result do begin
    x:= v1.x - v2.x; y := v1.y - v2.y; z := v1.z - v2.z;
    end;
  end;

function VectorMul(v1, v2: TD3DXVector3): TD3DXVector3;
begin
  with result do begin
    x := v1.x * v2.x; y := v1.y * v2.y; z := v1.z * v2.z;
    end;
  end;

function VectorMuls(v1: TD3DXVector3; s: single): TD3DXVector3;
begin
  with result do begin
    x := v1.x * s; y := v1.y * s; z := v1.z * s;
    end;
  end;

function VectorDiv(v1, v2: TD3DXVector3): TD3DXVector3;
begin
  with result do begin
    x := v1.x / v2.x; y := v1.y / v2.y; z := v1.z / v2.z;
    end;
  end;

function VectorDivs(v1: TD3DXVector3; s: single): TD3DXVector3;
begin
  with result do begin
    x := v1.x / s; y := v1.y / s; z := v1.z / s;
    end;
  end;

function VectorMagnitude(v1: TD3DXVector3): single;
begin
  result := sqrt(sqr(v1.x) + sqr(v1.y) + sqr(v1.z));
  end;

type
  WArray = array [1..30000] of word;
  PWord = ^WArray;

//-----------------------------------------------------------------------------
{ QUICKSORT sorts elements in the array "SList" of any record type of
  size "RSize" (in bytes) with indices between 0 and "Count"-1 (both inclusive).
  Note that the QuickSort procedure provides only an "interface" to the program.
  The actual processing takes place in the "Sort" procedure, which executes
  itself recursively.
  "Compare" is a callback function of type "TCompFunction" (see begin of unit).
  Algotrithm based on Borland pascal sample "qsort.pas",
  modified J. Rathlev, Nov. 2001 (rathlev@physik.uni-kiel.de) }
//-----------------------------------------------------------------------------
procedure QuickSort (var SList;
                     Count,RSize : Integer;
                     Compare     : TCompFunction);
var
  {the  following variables are global to "Sort" to spare memory in recursive calls }
  tmp,ns  : TByteArray;
  pl      : PByteArray;
  n,m     : integer;

  procedure Sort(nl,nr: Integer);
  var
    i,j   : integer;
  begin
    n:=RSize*((nl+nr) div 2);
    move (pl^[n],ns[0],RSize);
    i:=nl; j:=nr;
    repeat
      while Compare(pl^[RSize*i],ns[0]) do inc(i);
      while Compare(ns[0],pl^[RSize*j]) do dec(j);
      if i<=j then begin
        n:=RSize*i; m:=RSize*j;
        move (pl^[n],tmp[0],RSize);
        move (pl^[m],pl^[n],RSize);
        move (tmp[0],pl^[m],RSize);
        inc(i); dec(j);
        end;
      until i>j;
    if nl<j then Sort(nl,j);
    if i<nr then Sort(i,nr);
    end;

begin {QuickSort};
  SetLength(Tmp,RSize); SetLength(ns,RSize);
  pl:=@SList;
  Sort(0,Count-1);
  Tmp:=nil; ns:=nil;
  end;

//-----------------------------------------------------------------------------

end.
